#include "Field.hpp"

FieldCellCoord::FieldCellCoord( int x , int y ) 
    :
        x(x), y(y)
{
    /// Empty
}

FieldCell::FieldCell( int type , int64 turret )
    :
        type(type), turret(turret)
{
    
}

void FieldCell::setVisibleBy( int turret_id ){
    assert( turret_id >= 0 and turret_id < MAX_TURRET_COUNT );
    
    turret |= (1ULL << turret_id);
}

void FieldCell::setInvisibleBy( int turret_id ){
    assert( turret_id >= 0 and turret_id < MAX_TURRET_COUNT );
    
    turret &= ~(1ULL << turret_id );
}

bool FieldCell::isVisibleBy( int turret_id ) const{
    assert( turret_id >= 0 and turret_id < MAX_TURRET_COUNT );
    
    return turret & (1ULL << turret_id);
}

Field::Field( int rows , int cols , const char field[MAX_FIELD_SIZE][MAX_FIELD_SIZE] )
    :
        rows(rows), cols(cols), cell()
{
    cell.resize(rows);
    
    int start_x, start_y;
    
    for( int i = 0 ; i < rows ; i++ ){
        cell[i].resize( cols );
        
        for( int j = 0 ; j < cols ; j++ ){
            cell[i][j].turret = 0LL;
            
            switch( field[i][j] ){
                case 'S':
                case 'E':
                case '#':
                    cell[i][j].type = FIELD_PATH;
                    
                    //eprintf("Field[%d][%d] -> Path cell", i, j);
                    break;
                case '.':
                    cell[i][j].type = FIELD_FREE;
                    
                    //eprintf("Field[%d][%d] -> Free cell", i, j);
                    break;
                case 'x':
                    cell[i][j].type = FIELD_BLOCKED;
                    
                    //eprintf("Field[%d][%d] -> Blocked cell", i, j);
                    break;
            }
            
            if( field[i][j] == 'S' )
                start_x = i, start_y = j;
        }
    }
    
    int before_x = -1, before_y = -1;
    int x = start_x, y = start_y;
    
    while( field[x][y] != 'E' ){
        path.push_back( FieldCellCoord( x , y ) );
        
        for( int i = 0 ; i < DIRECTIONS ; i++ ){
            int new_x = x + DELTA_X[i];
            int new_y = y + DELTA_Y[i];
            
            if( isInside( new_x , new_y ) and !(new_x == before_x and new_y == before_y) ){
                if( cell[new_x][new_y].type == FIELD_PATH ){
                    before_x = x;
                    before_y = y;
                    
                    x = new_x;
                    y = new_y;
                    
                    break;
                }
            }
        }
    }
    
    /// Add the final cell of the path
    path.push_back( FieldCellCoord( x , y ) );
}

bool Field::isInside( int x , int y ) const{
    return x >= 0 and y >= 0 and x < rows and y < cols;
}

bool Field::placeTurret( int x , int y , int id ){
    assert( id >= 0 and id < MAX_TURRET_COUNT );
    
    assert( isInside( x , y ) );
    
    if( cell[x][y].type != FIELD_FREE )
        return false;
    
    cell[x][y] = FieldCell( FIELD_TURRET , id );
    
    return true;
}

bool Field::removeTurret( int x , int y ){
    if( cell[x][y].type != FIELD_TURRET )
        return false;
    
    /// Note that you would have to keep track of
    /// the turret list manually as the field is not
    /// really concerned with keeping track of them...
    cell[x][y] = FieldCell( FIELD_FREE , 0LL );
    
    return true;
}

/**
 * Marks (x,y) as "in the range of" turret_id
 * (x,y) needs to be of type FIELD_PATH
 * If turret_id < 0, marks as invisible
 */
bool Field::setVisibleBy( int x , int y , int turret_id ){
    assert( isInside( x , y ) );
    assert( turret_id >= -MAX_TURRET_COUNT - 1 and turret_id < MAX_TURRET_COUNT );
    
    if( cell[x][y].type != FIELD_PATH )
        return false;
    
    if( turret_id >= 0 )
        cell[x][y].setVisibleBy( turret_id );
    else
        cell[x][y].setInvisibleBy( -(turret_id + 1) );
    
    return true;
}

bool Field::setInvisibleBy( int x , int y , int turret_id ){
    assert( isInside( x , y ) );
    
    return setVisibleBy( x , y , -turret_id - 1 );
}

bool Field::isVisibleBy( int x , int y , int turret_id ) const{
    assert( isInside( x , y ) );
    
    return cell[x][y].isVisibleBy( turret_id );
}

bool Field::isFree( int x , int y ) const{
    assert( isInside( x , y ) );
    
    return cell[x][y].type == FIELD_FREE;
}

bool Field::isTurret( int x , int y ) const{
    assert( isInside( x , y ) );
    
    return cell[x][y].type == FIELD_TURRET;
}

bool Field::isPath( int x , int y ) const{
    assert( isInside( x , y ) );
    
    return cell[x][y].type == FIELD_PATH;
}

bool Field::isBlocked( int x , int y ) const{
    assert( isInside( x , y ) );
    
    return cell[x][y].type == FIELD_BLOCKED;
}

int Field::getTurretId( int x , int y ) const{
    assert( isTurret( x , y ) );
        
    return (int) cell[x][y].turret;
}

int64 Field::getTurretsSeeing( int x , int y ) const{
    assert( isPath( x , y ) );
    
    return cell[x][y].turret;
}
