#include "Game.hpp"

static int64 segmentBitwiseORAggregator( int64 a , int64 b ){
    return a | b;
}

static int64 segmentAssignUpdate( int64 a , int64 b ){
    return b;
}

static void nullWaveCompletedCallback( WaveInfo *result , Game *game ){
    //eprintf("Wave #%d completed", result->wave_id);
#if 0
    eprintf(" Mobs killed: %d (out of %d, rate = %.2lf%%)", 
            result->mobs_killed, result->mobs_total,
            100.0 * ((double) (result->mobs_killed) / (double) (result->mobs_total))
        );
#endif
	game->wave++;
}

/**
 * Try this algorithm where each turret shoots everything
 *  at once (till the mob is dead)
 * Combined w/ the current WaveRunner calling mobInRange just once / second
 * 
 * Try calling mobInRange more than once per second
 *  (i.e. shoot once per turret and call until mob is dead or mob health doesn't change)
 *
 *
 * Also try implementing splash damage
 */
void defaultMobInRangeForManyCalls( RunningMob *mob , PlacedTurret *turret , Game &game , WaveInfo *result , void *argument ){
    const ManyRunsLogicInfo *info = static_cast< ManyRunsLogicInfo* >( argument );
    
    const int mobType = mob->type->type;
    
    const int x = turret->place.x;
    const int y = turret->place.y;
    
    const int turretRate = turret->type->rate * game.rate_mod[mobType]->query( x , y , x + 1 , y + 1 );
    
    const double damage = game.damage_mod[mobType]->query( x , y , x + 1 , y + 1 ) * turret->type->affect[mobType];
    
    if( turret->type->rate >= 0 ){
        if( turret->fires >= turretRate )
            return;
    }else{
        /// Mass damage mode, fires irrespective of rate
        
        if( result->call_id > -turret->type->rate )
            return; /// hits just once
    }
    
		//if (damage > EPS)
    //printf("At %d:%d(%d) Mob(%d, HP = %.2lf) hit with %.2lf by %s (%d)\n", result->wave_id, result->length, result->call_id, mobType, mob->health, damage, turret->type->name.c_str(), turret->type->rate - turret->fires);
    
    /// Don't fire if too much waste?
    
    if( double_compare( damage , mob->health ) == -1 ){
        mob->health -= damage;
        
        result->total_damage_done += damage;
    }else{
        double waste = (damage - mob->health) / damage;
        
        if( double_compare( waste , info->maxWasteDamage[x][y] ) > 0 )
            return; /// too much waste
        
        result->total_waste_damage += (damage - mob->health);
        
        mob->health = 0;
        
        result->total_damage_done += damage;
    }
    
    turret->fires++;
}

void defaultMobInRangeForOneCall( RunningMob *mob , PlacedTurret *turret , Game &game , WaveInfo *result , void *argument ){
    /**
     * Check if it can fire at all
     */
    const int mobType = mob->type->type;
    
    const int x = turret->place.x;
    const int y = turret->place.y;
    
    const int turretRate = turret->type->rate * game.rate_mod[mobType]->query( x , y , x + 1 , y + 1 );
    
    if( turret->fires >= turretRate )
        return;
    
    int fires_left = turretRate - turret->fires;
    
    double damage = game.damage_mod[mobType]->query( x , y , x + 1 , y + 1 ) * turret->type->affect[mobType];
    
    int fires_needed = ceil( mob->health / damage );
    
    eprintf("Mob(%d, HP = %.2lf) in range of %s, damage done: %.2lf (hits = %d)",
                mobType, mob->health, turret->type->name.c_str(), std::min( fires_needed , fires_left ) * damage,
                std::min( fires_needed , fires_left )
        );
    
    if( fires_needed > fires_left ){
        mob->health -= damage * fires_left;
        
        turret->fires += fires_left;
        
        result->total_damage_done += damage * fires_left;
    }else{
        result->total_waste_damage += (fires_needed * damage) - mob->health;
        result->total_damage_done += fires_needed * damage;
        
        turret->fires += fires_needed;
        
        mob->health = 0;
    }
}

void defaultMobFellDead( RunningMob *mob , PlacedTurret *turret , Game &game , WaveInfo *result , void *argument ){
    //eprintf("Mob(%d) killed (lives now %d)", mob->type->type, game.lives);
    
    result->mobs_killed++;
    turret->kills++;
}

void defaultMobReachedTheEnd( RunningMob *mob , PlacedTurret *turret , Game &game , WaveInfo *result , void *argument ){
    //eprintf("Mob(%d) reached the end (HP: %.2lf)", mob->type->type, mob->health);
    
    game.lives--;
}

Game::Game( int rows , int cols , const char field[MAX_FIELD_SIZE][MAX_FIELD_SIZE] )
    :
        field( new Field( rows , cols , field ) ),
        visibilityTree( new Segmenttree( MAX_FIELD_SIZE * MAX_FIELD_SIZE , segmentBitwiseORAggregator , segmentAssignUpdate ) ),
        lives(1),
        kills(0),
        money(0ULL),
        wave(1),
        turrets( MAX_TURRET_COUNT , reinterpret_cast< PlacedTurret* >(NULL) ),
        _turret_mask( (0xFFFFFFFFFFFFFFFFULL) ), /// 16
        _turrets_built( 0 )
{
   for( int i = 0 ; i < MAX_MOB_TYPES ; i++ ){
        damage_mod[i] = new Quadtree( MAX_FIELD_SIZE , 1.0 );
        rate_mod[i] = new Quadtree( MAX_FIELD_SIZE , 1.0 );
   }
}

int Game::buyTurret( const Turret *turret , int x , int y ){
    if( _turrets_built == MAX_TURRET_COUNT )
        return -1;
    if( (int64) turret->buy_price > money )
        return -2;
    if( !field->isInside( x , y ) or !field->isFree( x , y ) )
        return -3;
    
    for( int i = 0 ; i < MAX_TURRET_COUNT ; i++ ){
        if( (POW2[i] & _turret_mask) ){
            //eprintf("Free turret position %d for `%s`", i, turret->name.c_str());
            
            _turret_mask &= ~POW2[i];
            
            turrets[i] = new PlacedTurret( x , y , turret );
            
            field->placeTurret( x , y , i );
            
            _turrets_built++;
            
            money -= turret->buy_price;
            
            int range = turret->range;
            
            if( turret->type == TURRET_BATTLE ){
                /**
                 * As modifier turrets affect the field, not
                 *  mobs, don't go around the mob path
                 */
                int len = (int) field->path.size();
                
                for( int j = 0 ; j < len ; j++ ){
                    int pathX = field->path[j].x;
                    int pathY = field->path[j].y;
                    
                    if( abs(pathX - x) <= range and abs(pathY - y) <= range ){
                        field->setVisibleBy( pathX , pathY , i );
                        visibilityTree->update( j , field->getTurretsSeeing( pathX , pathY ) );
                    }
                }
            }else{
                /**
                 * The base code for modifier turrets
                 */
                
                /// Calculate the indices
                int upper_left_x = std::max( 0 , x - range );
                int upper_left_y = std::max( 0 , y - range );
                
                int lower_right_x = std::min( field->rows , 1 + x + range );
                int lower_right_y = std::min( field->cols , 1 + y + range );
                
                for( int t = 0 ; t < MAX_MOB_TYPES ; t++ ){
                    if( double_compare( 0.0 , turret->affect[t] ) == 0 )
                        continue;
                    
                    Quadtree *tree;
                    
                    if( turret->type == TURRET_MOD_DAMAGE )
                        tree = damage_mod[t];
                    else if( turret->type == TURRET_MOD_RATE )
                        tree = rate_mod[t];
                    else{
                        assert( 0 and "Invald turret type?" );
                    }
                    
                    tree->update( upper_left_x , upper_left_y , lower_right_x , lower_right_y , turret->affect[t] );
                }
            }
            
            return i;
        }
    }
    
    
    return -1;
}

bool Game::sellTurret( int x , int y , bool remove_only ){
    if( !field->isTurret( x , y ) )
        return false;
    
    int turret_id = (int) field->getTurretId( x , y );
    
    assert( (POW2[turret_id] & _turret_mask) == 0ULL );
    
    assert( turret_id >= 0 and turret_id < MAX_TURRET_COUNT );
    
    /// Remove from the turret mask
    _turret_mask |= POW2[turret_id];
    
    /// Get the money
    if( !remove_only )
        money += turrets[turret_id]->type->sell_price;
    
    /// Update the path
    int len = (int) field->path.size();
    
    /**
     * Find a way not to repeat code ?
     */
    if( turrets[turret_id]->type->type == TURRET_BATTLE ){
        for( int i = 0 ; i < len ; i++ ){
            int pathX = field->path[i].x;
            int pathY = field->path[i].y;
            
            if( field->isVisibleBy( pathX , pathY , turret_id ) ){
                field->setInvisibleBy( pathX , pathY , turret_id );
                visibilityTree->update( i , field->getTurretsSeeing( pathX , pathY ) );
            }
        }
    }else{
        int upper_left_x = std::max( 0 , x - turrets[turret_id]->type->range );
        int upper_left_y = std::max( 0 , y - turrets[turret_id]->type->range );
        
        int lower_right_x = std::min( field->rows , x + turrets[turret_id]->type->range );
        int lower_right_y = std::min( field->cols , y + turrets[turret_id]->type->range );
        
        for( int t = 0 ; t < MAX_MOB_TYPES ; t++ ){
            Quadtree *tree;
            
            if( turrets[turret_id]->type->type == TURRET_MOD_DAMAGE )
                tree = damage_mod[t];
            else if( turrets[turret_id]->type->type == TURRET_MOD_RATE )
                tree = rate_mod[t];
            
            tree->update( upper_left_x , upper_left_y , lower_right_x , lower_right_y , -turrets[turret_id]->type->affect[t] );
        }
    }
    
    /// Free the memory taken by the PlacedTurret object
    delete turrets[turret_id];
    
    /// Restore the "no turret" indicator
    turrets[turret_id] = NULL;
    
    /// Free the cell
    field->removeTurret( x , y );
    
    /// Decrease the count of built turrets
    _turrets_built--;
    
    return true;
}

PlacedTurret *Game::getTurret( int turret_id ) const{
	//fprintf(stderr, "turret_id = %d\n", turret_id);
    assert( turret_id >= 0 and turret_id < MAX_TURRET_COUNT );
    
    return turrets[turret_id];
}

void Game::go( WaveRunner_Logic_Callback myWaveRunLogic , const std::vector< RunningMob* > &mobs ){
    go( myWaveRunLogic , mobs , nullWaveCompletedCallback , defaultMobInRangeForOneCall ,
        defaultMobFellDead , defaultMobReachedTheEnd , NULL );
}

void Game::go( WaveRunner_Logic_Callback myWaveRunLogic , const std::vector< RunningMob* > &mobs ,
                WaveRunner_Callback_Type myMobInRange ,
                WaveRunner_Callback_Type myMobFellDead ,
                WaveRunner_Callback_Type myMobReachedTheEnd ,
                void *argument ){
    go( myWaveRunLogic , mobs , nullWaveCompletedCallback , myMobInRange , myMobFellDead , myMobReachedTheEnd , argument );
}

void Game::go( WaveRunner_Logic_Callback myWaveRunLogic , const std::vector< RunningMob* > &mobs ,
                WaveCompleted_Callback_Type waveCompleted ,
                WaveRunner_Callback_Type myMobInRange ,
                WaveRunner_Callback_Type myMobFellDead ,
                WaveRunner_Callback_Type myMobReachedTheEnd ,
                void *argument ){
    WaveRunner *waveRunner = new WaveRunner;
    
    /**
     * Set up the callbacks
     */
    waveRunner->mobInRange = myMobInRange;
    waveRunner->mobFellDead = myMobFellDead;
    waveRunner->mobReachedTheEnd = myMobReachedTheEnd;
    
    waveRunner->waveRunLogic = myWaveRunLogic;
    
    WaveInfo *result = waveRunner->run( mobs , *this , argument );
    
    waveCompleted( result , this );
}

int Game::totalTurretsBuild() const{
    return _turrets_built;
}
