#include "WaveRunner.hpp"

struct DamageComparator{
    const Game *game;
    int damage_type;
    
    DamageComparator( const Game *game , int damage_type )
        : damage_type( damage_type ), game(game)
    {
        
    }
    
    inline bool operator () ( const int a_id , const int b_id ) const{
        const PlacedTurret *a = game->getTurret( a_id );
        const PlacedTurret *b = game->getTurret( b_id );
        
        if( a != NULL and b != NULL ){
            double aDamage = a->type->affect[damage_type] * game->damage_mod[damage_type]->query( a->place.x , a->place.y , a->place.x + 1 , a->place.y + 1);
            double bDamage = b->type->affect[damage_type] * game->damage_mod[damage_type]->query( b->place.x , b->place.y , b->place.x + 1 , b->place.y + 1);
            
            return double_compare( aDamage , bDamage ) > 0;
        }else if( a == NULL and b == NULL )
            return false;
        else if( a == NULL and b != NULL )
            return false;
        else if( a != NULL and b == NULL )
            return true;
    }
};

void waveRunLogicOneInRangeCall( WaveRunner *waveRunner , const std::vector< RunningMob* > &mobs , Game &game , WaveInfo *result , void *argument ){
    std::vector< int > mobs_left( mobs.size() );
    
    for( int i = 0 ; i < (int) mobs.size() ; i++ )
        mobs_left[i] = i;
    
    while( !mobs_left.empty() ){
        /**
         * Update the length of the wave
         */
        result->length++;
        result->call_id = 1; /// always, once in a second
        
        /**
         * Reset the fire count
         */
        for( int i = 0 ; i < MAX_TURRET_COUNT ; i++ ){
            int turret_id = i;
            
            PlacedTurret *turret = game.getTurret( turret_id );
            
            if( !turret )
                continue;
            
            turret->fires = 0;
        }
        
        int mobs_alive = 0;
        
        eprintf("got %d mobs", mobs_left.size());
        
        for( int m = 0 ; m < (int) mobs_left.size() ; m++ ){
            RunningMob *mob = mobs[ mobs_left[m] ];
            
            assert( double_compare( 0.0 , mob->health ) == -1 );
            
            bool stillAlive = true;
            
            for( int i = 0 ; i < MAX_TURRET_COUNT ; i++ ){
                int turret_id = i; /// was nthTurret[i]
                
                /**
                 * Add segment tree
                 */
                int endCell = std::min( MAX_FIELD_SIZE * MAX_FIELD_SIZE , mob->type->speed + mob->location );
                
                int64 visibility = game.visibilityTree->query( mob->location , endCell );
                
                if( (visibility & POW2[turret_id]) == 0ULL ) /// not visible by this turret
                    continue;
                
                /**
                 * Make automatic respect of tolerance[]
                 *  If waste damage would be waaay higher don't even call 
                 *  mobInRange()
                 */
                waveRunner->mobInRange( mob , game.getTurret( turret_id ) , game , result , argument );
                
                if( double_compare( 0.0 , mob->health ) >= 0 ){
                    waveRunner->mobFellDead( mob , game.getTurret( turret_id ) , game , result , argument );
                    
                    stillAlive = false;
                    
                    break;
                }
            }
            
            if( stillAlive ){
                /**
                 * See if reaches the end?
                 */
                mob->location += mob->type->speed;
                
                if( mob->location >= (int) game.field->path.size() ){
                    waveRunner->mobReachedTheEnd( mob , NULL , game , result , argument );
                }else{
                    /**
                     * Stays for another second..
                     */
                    std::swap( mobs_left[mobs_alive++] , mobs_left[m] );
                }
            }
        }
        
        /// Erase the dead
        mobs_left.erase( mobs_left.begin() + mobs_alive , mobs_left.end() );
    }
}

/**
 * Use perm[] by damage ?
 */

void waveRunLogicManyInRangeCalls( WaveRunner *waveRunner , const std::vector< RunningMob* > &mobs , Game &game , WaveInfo *result , void *argument ){
    static int order[MAX_MOB_TYPES][MAX_TURRET_COUNT];
    
    ManyRunsLogicInfo *info = static_cast< ManyRunsLogicInfo* >( argument );
    
    std::vector< int > mobs_left( mobs.size() );
    
    for( int i = 0 ; i < (int) mobs.size() ; i++ )
        mobs_left[i] = i;
    
    for( int i = 0 ; i < MAX_MOB_TYPES ; i++ ){
        if( info->isManualOrder[i] ){
            //eprintf("Manual ordering for mob type %d", i);
            
            for( int j = 0 ; j < MAX_TURRET_COUNT ; j++ )
                order[i][j] = info->manualOrder[i][j];
        }else{            
            //eprintf("Automatic ordering for mob type %d", i);
            
            for( int j = 0 ; j < MAX_TURRET_COUNT ; j++ )
                order[i][j] = j;
        }
    }
    
    for( int i = 0 ; i < MAX_MOB_TYPES ; i++ ){
        if( info->isManualOrder[i] )
            continue;
        
        std::stable_sort( order[i] , order[i] + MAX_TURRET_COUNT , DamageComparator( &game , i ) );
    }
    
#if 0 // spam
    for( int i = 0 ; i < MAX_TURRET_COUNT && game.getTurret( order[0][i] ) ; i++ ){
        eprintf("order[%d] -> %d (damage: %.2lf)", i, order[0][i], ( game.getTurret( order[0][i] ) ? game.getTurret( order[0][i] )->type->affect[0] : -1. ) );
    }
#endif
    
    while( !mobs_left.empty() ){
        /**
         * Update the length of the wave
         */
        result->length++;
        
        /**
         * Reset the fire count
         */
        for( int i = 0 ; i < MAX_TURRET_COUNT ; i++ ){
            int turret_id = i;
            
            PlacedTurret *turret = game.getTurret( turret_id );
            
            if( !turret )
                continue;
            
            turret->fires = 0;
        }
        
        int mobs_alive = 0;
        
        //eprintf("got %d mobs", mobs_left.size());
        
        for( int m = 0 ; m < (int) mobs_left.size() ; m++ ){
            RunningMob *mob = mobs[ mobs_left[m] ];
            
            double oldHealth = mob->health;
            
            assert( double_compare( 0.0 , mob->health ) == -1 );
            
            bool stillAlive = true;
            
            /**
             * Another way of handling the firing of the turrets, may be more natural
             */
            
            result->call_id = 0;
            
            do{
                result->call_id++;
                
                oldHealth = mob->health;
                
                for( int i = 0 ; i < MAX_TURRET_COUNT ; i++ ){
                    int turret_id = order[mob->type->type][i]; /// was nthTurret[i]
                    
                    if( game.getTurret( turret_id ) == NULL )
                        continue;
                    
                    /**
                     * Add segment tree, NOTE: this is stupid, actually you could iterate
                     *  over the set bits. ?? would it be faster though?
                     */
                    int endCell = std::min( MAX_FIELD_SIZE * MAX_FIELD_SIZE , mob->type->speed + mob->location );
                    
                    int64 visibility = game.visibilityTree->query( mob->location , endCell );
                    
                    if( (visibility & POW2[turret_id]) == 0ULL ) /// not visible by this turret
                        continue;
                    
                    /**
                     * Make automatic respect of tolerance[]
                     *  If waste damage would be waaay higher don't even call 
                     *  mobInRange()
                     * DONE! But implemented as a part of mobInRange()
                     */
                    waveRunner->mobInRange( mob , game.getTurret( turret_id ) , game , result , argument );
                    
                    if( double_compare( 0.0 , mob->health ) >= 0 ){
                        waveRunner->mobFellDead( mob , game.getTurret( turret_id ) , game , result , argument );
                        
                        stillAlive = false;
                        
                        break;
                    }
                }
                
            }while( stillAlive and double_compare( oldHealth , mob->health ) == 1 );
            
            if( stillAlive ){
                /**
                 * See if reaches the end?
                 */
                mob->location += mob->type->speed;
                
                if( mob->location >= (int) game.field->path.size() ){
                    waveRunner->mobReachedTheEnd( mob , NULL , game , result , argument );
                }else{
                    /**
                     * Stays for another second..
                     */
                    std::swap( mobs_left[mobs_alive++] , mobs_left[m] );
                }
            }
        }
        
        /// Erase the dead
        mobs_left.erase( mobs_left.begin() + mobs_alive , mobs_left.end() );
    }
}


WaveInfo *WaveRunner::run( const std::vector< RunningMob* > &mobs , Game &game , void *argument ){
    /**
     * Simulates a wave run
     */
    WaveInfo *result = new WaveInfo;
    
    result->wave_id = game.wave;
    result->length = 0;
    result->mobs_total = (int) mobs.size();
    result->mobs_killed = 0;
    result->total_damage_done = 0.0;
    result->total_waste_damage = 0.0;
    
    std::vector< int > mobs_left( mobs.size() );
    
    for( int i = 0 ; i < (int) mobs.size() ; i++ )
        mobs_left[i] = i;
    
    waveRunLogic( this , mobs , game , result , argument );
    
    return result;
}
